import datetime
import json
import logging
import math
import mimetypes
import os
import queue
import random
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from enum import Enum
from threading import BoundedSemaphore
from urllib.parse import urlparse

from django.db import connection, connections
from django.utils import dateparse
from . import decorators

logger = logging.getLogger(__name__)


class QueuedThreadExecutor(ThreadPoolExecutor):
    """
    Limits the queue size so all futures don't get created at once

    :param max_workers: Integer - the size of the thread pool
    """

    def __init__(self, max_workers=5, *args, **kwargs):
        super().__init__(max_workers=max_workers, *args, **kwargs)
        self._work_queue = queue.Queue(maxsize=max_workers)


class SerializerMixin:
    @property
    def as_dict(self):
        return json.loads(self.as_json)

    @property
    def as_json(self):
        return obj_to_json(self)


class BoundedThreadExecutor:
    """
    BoundedExecutor behaves as a ThreadPoolExecutor which will block on
    calls to submit() once the limit given as "bound" work items are queued for
    execution. Closes db connections on add_done_callback

    :param max_workers: Integer - the size of the thread pool
    :param bound: Integer - the maximum number of items in the work queue
    """

    def __init__(self, max_workers=5, bound=0):
        self.executor = ThreadPoolExecutor(max_workers=max_workers)
        self.semaphore = BoundedSemaphore(bound + max_workers)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.shutdown(wait=True)
        return False

    """See concurrent.futures.Executor#submit"""

    def submit(self, fn, *args, **kwargs):
        self.semaphore.acquire()
        try:
            future = self.executor.submit(decorators.db_conn_close(fn), *args, **kwargs)
        except Exception:
            self.semaphore.release()
            raise
        else:
            future.add_done_callback(lambda x: self.semaphore.release())
            return future

    """See concurrent.futures.Executor#shutdown"""

    def shutdown(self, wait=True):
        self.executor.shutdown(wait)


class ChoiceEnum(Enum):
    @classmethod
    def choices(cls):
        return [(tag.value, tag.name) for tag in cls]

    @classmethod
    def labels(cls):
        return [tag.name for tag in cls]

    @classmethod
    def by_label(cls, label):
        return [tag.value for tag in cls if label == tag.name][0]


class Pager:
    """
    Facilitates paging of generic apis.
    Page count starts at 1
    """

    def __init__(self, items_per_page=100, items_count=1):
        self.items_per_page = items_per_page
        self.items_count = items_count

    @property
    def items_count(self):
        return self._items_count

    @items_count.setter
    def items_count(self, items_count):
        if items_count < 1:
            raise ValueError('Items Count must be integer greater than zero')
        self._items_count = items_count

    @property
    def last_page_number(self):
        return math.ceil(self.items_count / self.items_per_page)

    @property
    def last_page_items_count(self):
        return self.items_on_page(self.last_page_number)

    def validate_page_number(self, page_number):
        if page_number < 1:
            raise ValueError('Page Number must be integer greater than zero')

        if page_number > self.last_page_number:
            raise ValueError(f'Page Number must be integer less than {self.last_page_number}.')

    def offset(self, page_number):
        self.validate_page_number(page_number)

        return (page_number - 1) * self.items_per_page

    def items_on_page(self, page_number):
        self.validate_page_number(page_number)

        if page_number == self.last_page_number:
            return self.items_count % self.items_per_page or self.items_per_page

        return self.items_per_page


def abbrev_str(s: str, max_length) -> str:
    return f'{s[:max_length]}...' if len(s) > max_length else s


def close_db_connections():
    connections.close_all()


def eye_catcher_line(msg):
    edge = '####'
    return f'{edge}  {msg}  {edge}'


def get_file_ext(uri):
    return os.path.splitext(urlparse(uri).path)[1]


def get_path(uri):
    return urlparse(uri).path


def get_mimetype(uri):
    return mimetypes.guess_type(get_path(uri))[0]


def get_weighted_item(items):
    return random.choices(population=[p[0] for p in items], weights=[p[1] for p in items], k=1)[0]


def obj_to_json(obj):
    return json.dumps(obj, default=lambda x: x.__dict__, separators=(',', ':'))


def pretty_dict(data):
    return json.dumps(
        data,
        default=lambda x: x.__dict__,
        sort_keys=True,
        indent=4
    )


def past_unix_timestamp_ss(unix_timestamp, **kwargs):
    dt = datetime.fromtimestamp(unix_timestamp) - timedelta(**kwargs)
    return int(dt.timestamp())


def utf_16_decode(msg):
    return msg.encode('utf-16', 'surrogatepass').decode('utf-16')


def iso8601_timestamp_to_unix(timestamp):
    return iso8601_timestamp_to_datetime(timestamp).timestamp() if timestamp else timestamp


def iso8601_timestamp_to_datetime(timestamp):
    return dateparse.parse_datetime(timestamp) if timestamp else timestamp


def exponential_decay(y_init, y_min, t, tau):
    """
    :param y_init:  Y-axis starting point
    :param y_min:   Y-axis minimum value
    :param t:       Time factor -> Higher = Lower y-axis value
    :param tau:     Decay factor -> Lower = faster decay

    :return:
    """
    y_scaled = y_init - y_min
    return (y_scaled * math.exp(-t / tau)) + y_min


def add_jitter(min_value, jitter, val):
    return max(min_value, val + random.uniform(-jitter, +jitter))


def dice_roll(chance):
    if chance > 1 or chance < 0:
        raise ValueError(f'{chance} is not a valid value')
    return random.random() <= chance



from contextlib import closing

import environ
from django.core.management.base import BaseCommand
from psycopg2 import connect
from psycopg2.errorcodes import DUPLICATE_DATABASE, DUPLICATE_OBJECT
from psycopg2 import errors
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

env = environ.Env()

TIMEOUT = 10

class Command(BaseCommand):
    help = 'Creates a new database'

    def add_arguments(self, parser):
        parser.add_argument(
            "--db-url",
            required=True,
            type=str,
            help="Database instance connection url",
        )
        parser.add_argument(
            "--db-name",
            required=True,
            type=str,
            help="New database name",
        )
        parser.add_argument(
            "--db-username",
            required=True,
            type=str,
            help="New database user username",
        )
        parser.add_argument(
            "--db-password",
            required=True,
            type=str,
            help="New database user password",
        )

    def handle(self, *args, **kwargs):
        self.stdout.write(self.style.SUCCESS('Starting db creation'))

        db_config = env.db_url_config(url=kwargs['db_url'])

        dbname = db_config['NAME']
        user = db_config['USER']
        password = db_config['PASSWORD']
        host = db_config['HOST']

        new_db_name = kwargs['db_name']
        new_db_username = kwargs['db_username']
        new_db_password = kwargs['db_password']

        with closing(connect(dbname=dbname, user=user, host=host, password=password, connect_timeout=TIMEOUT)) as con:
            con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
            with con.cursor() as cur:
                try:
                    cur.execute(f'CREATE DATABASE {new_db_name};')
                except errors.lookup(DUPLICATE_DATABASE):
                    self.stdout.write(self.style.WARNING(f'Database "{new_db_name}" already exists. Continuing...'))
                try:
                    cur.execute(f'CREATE USER {new_db_username} WITH ENCRYPTED PASSWORD \'{new_db_password}\';')
                except errors.lookup(DUPLICATE_OBJECT):
                    self.stdout.write(self.style.WARNING(f'User "{new_db_username}" already exists. Continuing...'))
                try:
                    cur.execute(f'GRANT ALL PRIVILEGES ON DATABASE {new_db_name} to {new_db_username};')
                    cur.execute(f'ALTER USER {new_db_username} CREATEDB;')
                except Exception:
                    self.stdout.write(self.style.ERROR(f'Error while granting user "{new_db_username}" privileges.'))
                    raise

                new_db_url = f'postgres://{new_db_username}:{new_db_password}@{host}:5432/{new_db_name}'
                self.stdout.write(self.style.SUCCESS((f'Successfully setup new DB "{new_db_name}". '
                                                      f'New db url: "{new_db_url}"')))



import logging
import time
from functools import wraps

from httpx import HTTPStatusError

from . import utils
from .exceptions import ApiError, NotAuthorized

logger = logging.getLogger(__name__)


def retry(retry_exceptions, tries=4, delay=1.5, backoff=2, jitter=0.1):
    """
    Retry calling the decorated function using an exponential backoff.

    Args:
        :param retry_exceptions: Exceptions to retry on before raise
        :param tries: Number of times to try (not retry) before giving up.
        :param delay: Initial delay between retries in seconds.
        :param backoff: Backoff multiplier (e.g. value of 2 will double the delay each retry).
        :param jitter: adds jitter to delay
    """
    def deco_func(func):

        @wraps(func)
        def wrapper_func(*args, **kwargs):

            m_tries, m_delay = tries, delay
            while m_tries > 1:
                try:
                    return func(*args, **kwargs)
                except retry_exceptions as e:
                    j_delay = utils.add_jitter(min_value=0, jitter=jitter, val=m_delay)
                    logger.warning(f'{type(e)} -> {str(e)} -> Retrying in {j_delay} seconds...')
                    time.sleep(j_delay)
                    m_tries -= 1
                    m_delay *= backoff

            logger.warning(f'Max tries reached: {tries}')
            return func(*args, **kwargs)

        return wrapper_func  # true decorator

    return deco_func


def api_error_check(func):

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HTTPStatusError as e:
            status_code = e.response.status_code
            if status_code in (408, 429) or status_code >= 500:     # Retryable status codes
                raise
            if status_code == 401:
                raise NotAuthorized(*args, **kwargs, **e.__dict__)

            raise ApiError(*args, **kwargs, **e.__dict__)

    return wrapper_func


def wrap_exceptions(raise_as):

    def deco_func(func):

        @wraps(func)
        def wrapper_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except raise_as:
                raise
            except Exception as e:
                raise raise_as(*args, **kwargs, **e.__dict__)

        return wrapper_func  # true decorator

    return deco_func


def db_conn_close(func):
    """
    Decorator that closes db connections after block executes

    :param func:
    :return:
    """

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        finally:
            utils.close_db_connections()

    return wrapper_func  # true decorator


def timing(func):
    """
    Decorator that times the execution of wrapped function
    """
    def wrap_func(*args, **kwargs):
        t1 = time.time()
        result = func(*args, **kwargs)
        t2 = time.time()
        logger.info(f'Function {func.__name__!r} executed in {(t2-t1):.4f}s')
        return result

    return wrap_func


def delay_fn(seconds=0, jitter=0.1):
    """
    Delays wrapped function

    :param seconds: Delay in seconds
    :param jitter: adds jitter to delay
    :return:
    """

    def deco_func(func):

        @wraps(func)
        def wrapper_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            finally:
                j_seconds = utils.add_jitter(min_value=0, jitter=jitter, val=seconds)
                logger.debug(f'Delaying for {j_seconds:.2f} seconds...')
                time.sleep(j_seconds)

        return wrapper_func  # true decorator

    return deco_func



from datetime import datetime, timezone
import functools
from unittest.mock import Mock

from django.db import connections
from django.test import TestCase, TransactionTestCase

from djstarter import models, utils


class QueuedExecutorTests(TransactionTestCase):
    def test_queued_thread_executor(self):
        def double(x): return 2 * x

        with utils.QueuedThreadExecutor(max_workers=5) as executor:

            future = executor.submit(double, 8)
            self.assertEquals(future.result(), 16)

            def raise_error(): raise ValueError
            future = executor.submit(raise_error)
            with self.assertRaises(ValueError):
                future.result()

        with self.assertRaises(RuntimeError):
            executor.submit(double, 8)


class BoundedExecutorTests(TransactionTestCase):
    def test_bounded_thread_executor(self):
        def double(x): return 2 * x

        with utils.BoundedThreadExecutor(max_workers=1) as executor:

            future = executor.submit(double, 8)
            self.assertEquals(future.result(), 16)

            def raise_error(): raise ValueError
            future = executor.submit(raise_error)
            with self.assertRaises(ValueError):
                future.result()

        with self.assertRaises(RuntimeError):
            executor.submit(double, 8)

    def test_db_connections_closed(self):
        func = Mock()

        def db_conn_func(label, f):
            f.result()
            models.ListItem.objects.create(label=label, value='val123')

            for c1 in connections.all():
                self.assertFalse(c1.connection.closed)

        with utils.BoundedThreadExecutor(max_workers=1) as executor:
            for _ in range(6):
                future = executor.submit(func)
                future.add_done_callback(functools.partial(db_conn_func, f'label{_}'))

        self.assertEquals(func.call_count, 6)


class PagerTests(TestCase):
    def test_pager_invalid_items_count(self):
        with self.assertRaises(ValueError):
            utils.Pager(items_count=0)

        with self.assertRaises(ValueError):
            utils.Pager(items_count=-333)

    def test_pager_page_number_less_than_1(self):
        pager = utils.Pager(items_count=999, items_per_page=111)

        with self.assertRaises(ValueError):
            pager.items_on_page(0)

        with self.assertRaises(ValueError):
            pager.items_on_page(-23)

    def test_pager_page_number_exceeds_last_page(self):
        pager = utils.Pager(items_count=999, items_per_page=111)

        with self.assertRaises(ValueError):
            pager.items_on_page(10)

        pager = utils.Pager(items_count=1525, items_per_page=25)

        with self.assertRaises(ValueError):
            pager.items_on_page(62)

    def test_pager_offset(self):
        pager = utils.Pager(items_count=1, items_per_page=1)

        self.assertEquals(pager.offset(1), 0)
        self.assertEquals(pager.offset(pager.last_page_number), 0)

        pager = utils.Pager(items_count=999, items_per_page=111)

        self.assertEquals(pager.offset(4), 333)
        self.assertEquals(pager.offset(7), 666)
        self.assertEquals(pager.offset(pager.last_page_number), 888)

        pager = utils.Pager(items_count=1525, items_per_page=23)

        self.assertEquals(pager.offset(4), 69)
        self.assertEquals(pager.offset(7), 138)
        self.assertEquals(pager.offset(pager.last_page_number), 1518)

    def test_pager_last_page_number(self):
        pager = utils.Pager(items_count=1, items_per_page=1)

        self.assertEquals(pager.last_page_number, 1)

        pager = utils.Pager(items_count=999, items_per_page=111)

        self.assertEquals(pager.last_page_number, 9)

        pager = utils.Pager(items_count=1525, items_per_page=23)

        self.assertEquals(pager.last_page_number, 67)

    def test_pager_last_page_items_count(self):
        pager = utils.Pager(items_count=1, items_per_page=1)

        self.assertEquals(pager.last_page_items_count, 1)

        pager = utils.Pager(items_count=999, items_per_page=111)

        self.assertEquals(pager.last_page_items_count, 111)

        pager = utils.Pager(items_count=1525, items_per_page=23)

        self.assertEquals(pager.last_page_items_count, 7)

    def test_pager_items_on_page(self):
        pager = utils.Pager(items_count=1, items_per_page=1)

        self.assertEquals(pager.items_on_page(1), 1)
        self.assertEquals(pager.items_on_page(pager.last_page_number), 1)

        pager = utils.Pager(items_count=999, items_per_page=111)

        self.assertEquals(pager.items_on_page(2), 111)
        self.assertEquals(pager.items_on_page(7), 111)
        self.assertEquals(pager.items_on_page(pager.last_page_number), 111)

        pager = utils.Pager(items_count=1525, items_per_page=23)

        self.assertEquals(pager.items_on_page(3), 23)
        self.assertEquals(pager.items_on_page(11), 23)
        self.assertEquals(pager.items_on_page(pager.last_page_number), 7)


class UtilTests(TestCase):
    """
    Utility Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.test_str = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit.'

    def test_choice_enum(self):
        class Statuses(utils.ChoiceEnum):
            INIT = 0
            RUNNING = 1
            ERROR = 2

        self.assertListEqual(Statuses.choices(), [(0, 'INIT'), (1, 'RUNNING'), (2, 'ERROR')])
        self.assertListEqual(Statuses.labels(), ['INIT', 'RUNNING', 'ERROR'])
        self.assertEquals(Statuses.by_label('RUNNING'), 1)

    def test_abbrev_str(self):
        self.assertEquals(utils.abbrev_str(self.test_str, max_length=10), 'Lorem ipsu...')
        self.assertEquals(utils.abbrev_str(self.test_str, max_length=100), self.test_str)

    def test_close_db_connections(self):
        # Open a DB connection
        models.ListItem.objects.create(label='test123', value='val123')

        for c1 in connections.all():
            self.assertFalse(c1.connection.closed)

        utils.close_db_connections()

        for c2 in connections.all():
            self.assertTrue(c2.connection.closed)

    def test_eye_catcher_line(self):
        self.assertEquals(utils.eye_catcher_line(self.test_str), f'####  {self.test_str}  ####')

    def test_get_weighted_item(self):
        items = [('abc', .25), ('def', .25), ('jkl', .5)]
        item = utils.get_weighted_item(items)
        self.assertIsNotNone(item)

    def test_obj_to_json(self):
        class Main1:
            class Sub1:
                def __init__(self, test_1):
                    self.test_1 = test_1

            def __init__(self, arg_1, arg_2, arg_3):
                self.arg_1 = arg_1
                self.arg_2 = arg_2
                self.arg_3 = self.Sub1(arg_3)

        obj = Main1(arg_1='value_1', arg_2='value_2', arg_3='value_3')
        self.assertEquals(
            utils.obj_to_json(obj),
            '{"arg_1":"value_1","arg_2":"value_2","arg_3":{"test_1":"value_3"}}'
        )

    def test_pretty_dict(self):
        obj = {
            't_1': 'val_1',
            't_8': {
                'l_3': 'val_3',
            },
            'test_7': {
                'l_19': 'val_30',
                'l_12': 'val_17'
            },
        }
        self.assertEquals(
            utils.pretty_dict(obj),
            ("{\n    \"t_1\": \"val_1\",\n    \"t_8\": {\n        \"l_3\": \"val_3\"\n    },\n    \"test_7\": {\n"
             "        \"l_12\": \"val_17\",\n        \"l_19\": \"val_30\"\n    }\n}")
        )

    def test_past_unix_timestamp_ss(self):
        start = 1635483705

        timestamp = utils.past_unix_timestamp_ss(unix_timestamp=start, hours=6)
        self.assertIsInstance(timestamp, int)
        self.assertEquals(timestamp, 1635462105)

        timestamp = utils.past_unix_timestamp_ss(unix_timestamp=start, days=2)
        self.assertEquals(timestamp, 1635310905)

        timestamp = utils.past_unix_timestamp_ss(unix_timestamp=start, days=7)
        self.assertEquals(timestamp, 1634878905)

        timestamp = utils.past_unix_timestamp_ss(unix_timestamp=start, days=28)
        self.assertEquals(timestamp, 1633064505)

        timestamp = utils.past_unix_timestamp_ss(unix_timestamp=start, days=1826)   # 5 years
        self.assertEquals(timestamp, 1477717305)

    def test_url_file_ext(self):
        test_url_1 = ('https://external-preview.redd.it/mPeN_GXOIAZmu1qO_gpQNiItIqiqbPL-Np_Hncp7_Mg.png?format=pjpg'
                      '&auto=webp&s=d51a4b61d82bcb2415de35e1ef35d986d2aad8d6')
        self.assertEquals(utils.get_file_ext(test_url_1), '.png')

        test_url_2 = 'https://imgur.com/V6lKWen.jpg'
        self.assertEquals(utils.get_file_ext(test_url_2), '.jpg')

    def test_get_path(self):
        test_url_1 = ('https://external-preview.redd.it/mPeN_GXOIAZmu1qO_gpQNiItIqiqbPL-Np_Hncp7_Mg.png?format=pjpg'
                      '&auto=webp&s=d51a4b61d82bcb2415de35e1ef35d986d2aad8d6')
        self.assertEquals(utils.get_path(test_url_1), '/mPeN_GXOIAZmu1qO_gpQNiItIqiqbPL-Np_Hncp7_Mg.png')

        test_url_2 = 'https://imgur.com/V6lKWen.jpg'
        self.assertEquals(utils.get_path(test_url_2), '/V6lKWen.jpg')

    def test_get_mimetype(self):
        test_url_1 = ('https://external-preview.redd.it/mPeN_GXOIAZmu1qO_gpQNiItIqiqbPL-Np_Hncp7_Mg.png?format=pjpg'
                      '&auto=webp&s=d51a4b61d82bcb2415de35e1ef35d986d2aad8d6')
        self.assertEquals(utils.get_mimetype(test_url_1), 'image/png')

        test_url_2 = 'https://imgur.com/V6lKWen.jpg'
        self.assertEquals(utils.get_mimetype(test_url_2), 'image/jpeg')

    def test_utf_16_decode(self):
        test_msg = '\ud83d\udd25 Whale swimming casually among a group of surfers unaware of his presence.'

        val = utils.utf_16_decode(test_msg)
        self.assertEquals(val, '🔥 Whale swimming casually among a group of surfers unaware of his presence.')

        test_msg = '\u201cAlright, is it going to be a Palkia?\u201d'
        val = utils.utf_16_decode(test_msg)

        self.assertEquals(val, '“Alright, is it going to be a Palkia?”')

    def test_exponential_decay(self):
        hours = 18
        seconds = hours * 60 * 60

        init_y = 1.0
        lower_bound = .23
        tau = 20000

        val = utils.exponential_decay(t=0, y_init=init_y, tau=tau, y_min=lower_bound)
        self.assertEquals(val, 1.0)

        val = utils.exponential_decay(t=1000, y_init=init_y, tau=tau, y_min=lower_bound)
        self.assertAlmostEquals(val, 0.9624466568655498, places=8)

        val = utils.exponential_decay(t=2000, y_init=init_y, tau=tau, y_min=lower_bound)
        self.assertAlmostEquals(val, 0.9267248118876888, places=8)

        val = utils.exponential_decay(t=10000, y_init=init_y, tau=tau, y_min=lower_bound)
        self.assertAlmostEquals(val, 0.6970286079787278, places=8)

        val = utils.exponential_decay(t=50000, y_init=init_y, tau=tau, y_min=lower_bound)
        self.assertAlmostEquals(val, 0.29320544894040207, places=8)

        val = utils.exponential_decay(t=seconds, y_init=init_y, tau=tau, y_min=lower_bound)
        self.assertAlmostEquals(val, 0.2601561992262201, places=8)

    def test_iso8601_timestamp_to_unix(self):
        self.assertEquals(utils.iso8601_timestamp_to_unix(''), '')
        self.assertEquals(utils.iso8601_timestamp_to_unix('2021-12-24T22:00:49.000Z'), 1640383249)
        self.assertEquals(utils.iso8601_timestamp_to_unix('2021-12-23T07:25:56.000Z'), 1640244356)

    def test_iso8601_timestamp_to_datetime(self):
        self.assertEquals(utils.iso8601_timestamp_to_datetime(''), '')

        dt = utils.iso8601_timestamp_to_datetime('2021-12-24T22:00:49.000Z')
        self.assertEquals(dt, datetime(2021, 12, 24, 22, 00, 49, tzinfo=timezone.utc))

        dt = utils.iso8601_timestamp_to_datetime('2021-12-23T07:25:56.000Z')
        self.assertEquals(dt, datetime(2021, 12, 23, 7, 25, 56, tzinfo=timezone.utc))

    def test_add_jitter(self):
        self.assertTrue(.5 <= utils.add_jitter(min_value=0, jitter=.5, val=1.0) <= 1.5)
        self.assertTrue(11 <= utils.add_jitter(min_value=11, jitter=20, val=15.5) <= 35.5)
        self.assertTrue(0.1437 <= utils.add_jitter(min_value=.07, jitter=.0123, val=.156) <= 0.1683)

    def test_dice_roll(self):
        is_true = utils.dice_roll(chance=1)
        self.assertTrue(is_true)

        is_false = utils.dice_roll(chance=0)
        self.assertFalse(is_false)

        with self.assertRaises(ValueError):
            utils.dice_roll(chance=2.05)

        with self.assertRaises(ValueError):
            utils.dice_roll(chance=-.09)




import csv
import itertools
import time
from urllib.parse import parse_qs

from django.conf.locale.en import formats as en_formats
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.http import HttpResponse

from . import models

en_formats.DATETIME_FORMAT = "M d y H:i"


class ExportCsvMixin:
    @classmethod
    def field_values(cls, obj, field_names, query_fields=None):
        vals = [getattr(obj, field) for field in field_names]
        if query_fields:
            queries = [q.replace('?', '') for q in vals if '=' in str(q)]
            parts = [[v[0] for k, v in parse_qs(q).items() if k in query_fields] for q in queries]
            return list(itertools.chain.from_iterable(parts))
        return vals

    def export_as_csv(self, request, queryset, query_fields=None):
        meta = self.model._meta
        model_name = str(meta).split('.')[1]
        field_names = [field.name for field in meta.fields]

        rows = [self.field_values(obj, field_names, query_fields) for obj in queryset]
        return self.write_to_csv(model_name, query_fields or field_names, rows)

    @classmethod
    def write_to_csv(cls, filename, field_names, rows):
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = f'attachment; filename={filename}_{time.strftime("%Y%m%d")}.csv'
        writer = csv.writer(response, quoting=csv.QUOTE_NONNUMERIC)
        writer.writerow(field_names)
        for row in rows:
            writer.writerow(row)

        return response

    export_as_csv.short_description = "Export Selected To CSV"


class BaseAdmin(admin.ModelAdmin, ExportCsvMixin):
    read_only_exclude = ()
    show_full_result_count = False

    def get_readonly_fields(self, request, obj=None):
        return [f.name for f in self.model._meta.fields if f.name not in self.read_only_exclude]

    def bulk_delete(self, request, queryset):
        queryset.delete()

    bulk_delete.short_description = 'Bulk Delete'

    actions = ('export_as_csv', bulk_delete, )


@admin.register(models.AuthUser)
class AuthUserAdmin(UserAdmin):
    list_display = ['date_joined', 'username', 'email', 'is_staff', 'is_superuser']
    list_filter = ('is_staff', 'is_superuser')
    ordering = ['-username']
    search_fields = ['username', ]

    show_full_result_count = False

    def bulk_delete(self, request, queryset):
        queryset.delete()

    bulk_delete.short_description = 'Bulk Delete'

    def unlock_user(self, request, queryset):
        queryset.update(is_locked=False, locked_at=None)

    unlock_user.short_description = 'Unlock User'

    actions = ['export_as_csv', bulk_delete, unlock_user, ]



from unittest import mock

from django.test import TestCase
from httpx import HTTPStatusError, NetworkError, Request, Response, codes

from djstarter.clients import RetryClient
from djstarter.exceptions import ApiError


class Http2ClientTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        cls.mocked_sleep = mock.patch('time.sleep', return_value=None).start()

    @mock.patch.object(RetryClient, '_send_handling_auth')
    def test_ok(self, mock_send):
        mock_send.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            text='ok'
        )
        with RetryClient() as h_client:
            h_client.get('http://example.com')
            self.assertEquals(mock_send.call_count, 1)

    @mock.patch.object(RetryClient, '_send_handling_auth')
    def test_retry(self, mock_send):
        mock_send.side_effect = NetworkError(message='')
        with self.assertRaises(NetworkError):
            with RetryClient() as h_client:
                h_client.get('https://example.com')
        self.assertEquals(mock_send.call_count, 4)

    @mock.patch.object(RetryClient, '_send_handling_auth')
    def test_raise_api_error(self, mock_send):
        response = Response(
            request=self.request,
            status_code=codes.NOT_FOUND,
        )
        mock_send.side_effect = HTTPStatusError(message='test_123', request=self.request, response=response)
        with self.assertRaises(ApiError):
            with RetryClient() as h_client:
                h_client.get('https://example.com')
        self.assertEquals(mock_send.call_count, 1)



# Generated by Django 3.2.12 on 2022-02-21 13:46

import django.core.validators
from django.db import migrations, models
import django.utils.timezone
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('djstarter', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Task',
            fields=[
                ('oid', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('scheduled_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('started_at', models.DateTimeField(blank=True, null=True)),
                ('ended_at', models.DateTimeField(blank=True, null=True)),
                ('status', models.IntegerField(choices=[(0, 'INIT'), (1, 'RUNNING'), (2, 'ERROR'), (3, 'FINISHED'), (4, 'STOPPING'), (5, 'STOPPED'), (6, 'CANCELLED'), (7, 'SCHEDULED')], db_index=True, default=0)),
                ('action_count', models.IntegerField(default=0)),
                ('target_action_count', models.IntegerField(default=1, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(200)])),
                ('thread_count', models.IntegerField(default=1, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(10)])),
                ('parameters', models.JSONField(default=dict)),
                ('results', models.JSONField(default=dict)),
            ],
            options={
                'db_table': 'core_task',
                'ordering': ['-created'],
            },
        ),
    ]



from django.test import TestCase

from djstarter.models import AuthUser, ListItem


class AuthManagerTests(TestCase):

    """
    AuthUser Manager Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.auth_user_data = {
            'username': 'username_1',
            'password': 'password_1',
            'is_superuser': True,
        }

    def test_by_id(self):
        self.assertEquals(AuthUser.objects.count(), 0)
        with self.assertRaises(AuthUser.DoesNotExist):
            AuthUser.objects.by_id('68992415-fd80-49c4-8dbf-4b0d1e5bafb9')
        auth_user = AuthUser.objects.create_user(**self.auth_user_data)
        auth_user = AuthUser.objects.by_id(auth_user.id)
        self.assertIsInstance(auth_user, AuthUser)

    def test_get_service_user(self):
        user = AuthUser.objects.get_service_user()
        self.assertIsNone(user)

        AuthUser.objects.create_user(**self.auth_user_data)
        user = AuthUser.objects.get_service_user()
        self.assertIsInstance(user, AuthUser)


class ListItemManagerTests(TestCase):

    """
    List Item Manager Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.list_item_data = {
            'group': 'group_1',
            'label': 'label_1',
            'value': 'value_1',
        }

    def test_list_item_str(self):
        list_item = ListItem.objects.create(**self.list_item_data)
        val = ListItem.objects.get_value(group=list_item.group, label=list_item.label)

        self.assertEquals(val, 'value_1')

    def test_get_by_oid(self):
        list_item_1 = ListItem.objects.create(**self.list_item_data)
        val = ListItem.objects.get_by_oid(oid=list_item_1.oid)

        self.assertEquals(list_item_1, val)



import logging

from httpx import Client, HTTPStatusError, TimeoutException, NetworkError, ProxyError

from djstarter import decorators

logger = logging.getLogger(__name__)

RETRY_EXCEPTIONS = (
    HTTPStatusError,
    TimeoutException,
    NetworkError,
    ProxyError
)


class RetryClient(Client):
    """
    Http/2 Client
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            follow_redirects=True,
            *args,
            **kwargs
        )

    @decorators.retry(retry_exceptions=RETRY_EXCEPTIONS)
    @decorators.api_error_check
    def send(self, *args, **kwargs):
        return super().send(*args, **kwargs)



from django.contrib.auth.models import UserManager
from django.db import models
from django.db.models import Q


class BaseManager(models.Manager):

    def get_by_oid(self, oid):
        return super().get_queryset().get(oid=oid)

      
class AuthUserManager(UserManager):
    def by_id(self, _id):
        q = Q(id=_id)
        return super().get_queryset().filter(q).get()

    def get_service_user(self):
        q = Q(is_superuser=True)
        return super().get_queryset().filter(q).first()



class ListItemManager(BaseManager):

    def get_value(self, group, label):
        q = Q(group=group, label=label)
        li_obj = super().get_queryset().filter(q).only('value').first()
        return li_obj.value if li_obj else None



import datetime

from django.test import TestCase
from django.utils import timezone

from djstarter.models import Task, ListItem
from djstarter.exceptions import AppError


class ListItemTests(TestCase):

    """
    List Item Model Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.group = 'group_1'
        cls.label = 'label_1'
        cls.value = 'value_1'

    def test_list_item_str(self):
        list_item = ListItem(group=self.group, label=self.label, value=self.value)
        self.assertEquals(str(list_item), f'ListItem: {self.label} / {self.value}')


class TaskTests(TestCase):
    """
    Task Tests
    """

    def test_actions_remaining(self):
        task = Task.objects.create(action_count=16, target_action_count=52)
        self.assertEquals(task.actions_remaining, 36)

    def test_percent_done(self):
        task = Task.objects.create(target_action_count=0)
        self.assertEquals(task.percent_done, 0)
        task = Task.objects.create(target_action_count=15)
        task.increment_action_count(increment_by=3)
        self.assertEquals(task.percent_done, 20)

    def test_time_elapsed(self):
        ms_offset = 982
        started_at = timezone.now()
        ended_at = started_at + datetime.timedelta(milliseconds=ms_offset)
        task = Task.objects.create(
            target_action_count=15
        )
        self.assertIsNone(task.time_elapsed)

        task.set_started_at(started_at=started_at)
        self.assertIsNone(task.time_elapsed)

        task.set_ended_at(ended_at=ended_at)
        self.assertEquals(task.time_elapsed, f'0:00:00.{ms_offset}000')

    def test_set_target_action_count(self):
        task = Task.objects.create(target_action_count=5)
        self.assertEquals(task.target_action_count, 5)
        task.set_target_action_count(12)
        self.assertEquals(task.target_action_count, 12)

    def test_set_thread_count(self):
        task = Task.objects.create(thread_count=17)
        self.assertEquals(task.thread_count, 17)

    def test_did_reach_target_action_count(self):
        task = Task.objects.create(target_action_count=15)
        task.increment_action_count(increment_by=4)
        self.assertFalse(task.did_reach_target_action_count)
        task.increment_action_count(increment_by=11)
        self.assertTrue(task.did_reach_target_action_count)

    def test_task_is_waiting(self):
        task = Task.objects.create()
        task.set_init_status()
        self.assertTrue(task.is_waiting)

    def test_task_is_running(self):
        task = Task.objects.create()
        self.assertFalse(task.is_running)
        task.start()
        self.assertTrue(task.is_running)

    def test_task_has_error(self):
        task = Task.objects.create()
        self.assertFalse(task.has_error)
        task.set_error_shutdown_status(e=AppError)
        self.assertTrue(task.has_error)

    def test_task_is_finished(self):
        task = Task.objects.create()
        self.assertFalse(task.is_finished)

        task.set_error_shutdown_status(e=ValueError)
        task.set_finished()
        self.assertFalse(task.is_finished)

        task.set_cancelled_status()
        task.set_finished()
        self.assertTrue(task.is_finished)

    def test_task_is_stopping(self):
        task = Task.objects.create()
        self.assertFalse(task.is_stopping)
        task.set_stopping_status()
        self.assertTrue(task.is_stopping)

    def test_task_is_cancelled(self):
        task = Task.objects.create()
        self.assertFalse(task.is_cancelled)
        task.set_cancelled_status()
        self.assertTrue(task.is_cancelled)

    def test_pretty_status(self):
        task = Task.objects.create()
        task.set_finished()
        self.assertEquals(task.pretty_status, 'finished')



import logging
import uuid

from django.contrib.auth.models import AbstractUser
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils import timezone

from . import managers, utils

logger = logging.getLogger(__name__)


class BaseModel(models.Model):
    oid = models.UUIDField(primary_key=True, default=uuid.uuid4)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

    @property
    def oid_str(self):
        return str(self.oid)


class BaseTask(BaseModel):
    """
    Base task model
    """

    class Statuses(utils.ChoiceEnum):
        INIT = 0
        RUNNING = 1
        ERROR = 2
        FINISHED = 3
        STOPPING = 4
        STOPPED = 5
        CANCELLED = 6
        SCHEDULED = 7

    scheduled_at = models.DateTimeField(default=timezone.now)
    started_at = models.DateTimeField(null=True, blank=True)
    ended_at = models.DateTimeField(null=True, blank=True)
    status = models.IntegerField(default=Statuses.INIT.value, choices=Statuses.choices(), db_index=True)
    action_count = models.IntegerField(default=0)
    target_action_count = models.IntegerField(default=1, validators=[MinValueValidator(1), MaxValueValidator(200)], )
    thread_count = models.IntegerField(default=1, validators=[MinValueValidator(1), MaxValueValidator(10)], )
    parameters = models.JSONField(null=False, default=dict)
    results = models.JSONField(null=False, default=dict)

    class Meta:
        abstract = True

    @property
    def actions_remaining(self):
        return self.target_action_count - self.action_count

    @property
    def percent_done(self):
        try:
            return int(min(self.action_count / self.target_action_count * 100, 100))
        except ZeroDivisionError:
            return 0

    @property
    def time_elapsed(self):
        if not (self.started_at and self.ended_at):
            return None
        return str(self.ended_at - self.started_at)

    @property
    def did_reach_target_action_count(self):
        return self.action_count >= self.target_action_count

    # Status Checks

    @property
    def is_waiting(self):
        return self.status in (self.Statuses.INIT.value, self.Statuses.SCHEDULED.value)

    @property
    def is_running(self):
        return self.status == self.Statuses.RUNNING.value

    @property
    def has_error(self):
        return self.status == self.Statuses.ERROR.value

    @property
    def is_finished(self):
        return self.status == self.Statuses.FINISHED.value

    @property
    def is_stopping(self):
        return self.status == self.Statuses.STOPPING.value

    @property
    def is_cancelled(self):
        return self.status == self.Statuses.CANCELLED.value

    # Field Prettifiers

    @property
    def pretty_status(self):
        return self.get_status_display().lower()

    @property
    def pretty_parameters(self):
        return utils.pretty_dict(self.parameters)   # pragma: no cover

    @property
    def pretty_results(self):
        return utils.pretty_dict(self.results)  # pragma: no cover

    def set_target_action_count(self, target_action_count):
        self.target_action_count = target_action_count
        self.save()

    def set_thread_count(self, thread_count):
        self.thread_count = thread_count
        self.save()

    def set_started_at(self, started_at=timezone.now):
        self.started_at = started_at
        self.save()

    def set_ended_at(self, ended_at=timezone.now):
        self.ended_at = ended_at
        self.save()

    def increment_action_count(self, increment_by=1):
        self.action_count += increment_by
        if self.did_reach_target_action_count:
            self.set_stopping_status()
            logger.info(f'{self} reached target action count: {self.action_count}')
        self.save()
        logger.info(f'{self} action count: {self.action_count}')

    def set_init_status(self):
        self.status = self.Statuses.INIT.value
        self.results = dict()
        self.save()

    def set_error_shutdown_status(self, e):
        self.results['error'] = str(e)
        self.status = self.Statuses.ERROR.value
        self.save()
        logger.exception(f'{self} Error Event: {type(e)}')

    def set_finished(self, results=None):
        if not self.has_error:
            self.status = self.Statuses.FINISHED.value
        self.results.update(results or dict())
        self.ended_at = timezone.now()
        self.save()

    def set_stopping_status(self):
        self.status = self.Statuses.STOPPING.value
        self.save()

    def set_cancelled_status(self):
        self.status = self.Statuses.CANCELLED.value
        self.save()

    def start(self):
        self.status = self.Statuses.RUNNING.value
        self.started_at = timezone.now()
        self.save()


class Task(BaseTask):

    class Meta:
        db_table = 'core_task'
        ordering = ['-created']
        app_label = 'djstarter'


class AuthUser(AbstractUser):
    objects = managers.AuthUserManager()
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    class Meta:
        db_table = 'core_authuser'
        ordering = ['-username']
        app_label = 'djstarter'


class ListItem(BaseModel):
    group = models.CharField(max_length=20)
    label = models.CharField(max_length=64)
    value = models.CharField(max_length=64)

    objects = managers.ListItemManager()

    class Meta:
        db_table = 'core_listitem'
        ordering = ['-created']
        app_label = 'djstarter'
        constraints = [
            models.UniqueConstraint(fields=['group', 'label'], name='list_item_unique'),
        ]

    def __str__(self):
        return f'ListItem: {self.label} / {self.value}'



import os

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Print Environmental Variables'

    def handle(self, *args, **kwargs):
        self.stdout.write(self.style.WARNING('### START Environmental Settings ####'))
        for k, v in os.environ.items():
            self.stdout.write(self.style.SUCCESS(f'{k}={v}'))
        self.stdout.write(self.style.WARNING('### END Environmental Settings ####'))



class Error(Exception):
    """There was an ambiguous exception that occurred while handling your
    request.
        """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, kwargs)
        self._args = args
        self._kwargs = kwargs

    def __str__(self):
        return f'{self.__class__.__name__}\nargs: {self._args}\nkwargs: {self._kwargs}'


class AppError(Error):
    """All unrecoverable Exceptions inherit from this class"""


class ApiError(Error):
    """All Api Exceptions inherit from this class"""


class NotAuthorized(ApiError):
    """All Http Status 401 Exceptions inherit from this class"""



from http import HTTPStatus
from unittest.mock import Mock

import httpx
from django.test import TestCase
from httpx import Request, Response

from djstarter import decorators, exceptions


class RetryTests(TestCase):
    """
    Decorator Tests
    """

    def test_retry(self):
        tries = 5
        func = Mock(side_effect=exceptions.ApiError("Test"))
        decorated_function = decorators.retry(
            retry_exceptions=(exceptions.ApiError,),
            tries=tries,
            delay=0,
            backoff=1,
            jitter=0,
        )(func)
        with self.assertRaises(exceptions.ApiError):
            decorated_function()
        self.assertEquals(func.call_count, tries)


class APIErrorCheckTests(TestCase):

    def test_check_response_400(self):
        request = Request(url='', method='')
        func = Mock(side_effect=httpx.HTTPStatusError(
            request=request,
            response=Response(
                request=request,
                status_code=HTTPStatus.BAD_REQUEST,
            ),
            message='test'
        ))
        decorated_function = decorators.api_error_check(func)
        with self.assertRaises(exceptions.ApiError):
            decorated_function()

    def test_check_response_502(self):
        request = Request(url='', method='')
        func = Mock(side_effect=httpx.HTTPStatusError(
            request=request,
            response=Response(
                request=request,
                status_code=HTTPStatus.BAD_GATEWAY,
            ),
            message='test'
        ))
        decorated_function = decorators.api_error_check(func)
        with self.assertRaises(httpx.HTTPStatusError):
            decorated_function()


class WrapExceptions(TestCase):
    def test_ok(self):
        func = Mock()
        decorated_function = decorators.wrap_exceptions(raise_as=BlockingIOError)
        decorated_function(func)()
        self.assertTrue(func.called)

    def test_pass_thru(self):
        func = Mock(side_effect=BlockingIOError)
        decorated_function = decorators.wrap_exceptions(raise_as=OSError)
        with self.assertRaises(BlockingIOError):
            decorated_function(func)()

    def test_runtime_error(self):
        func = Mock(side_effect=RuntimeError)
        decorated_function = decorators.wrap_exceptions(raise_as=OSError)
        with self.assertRaises(OSError):
            decorated_function(func)()


class DBConnCloseExceptions(TestCase):

    def test_db_conn_close(self):
        func = Mock()
        decorators.db_conn_close(func)()
        self.assertEquals(func.call_count, 1)


class TimingTests(TestCase):

    def test_timing(self):
        func = Mock(return_value=True)
        func.__name__ = 'test_func'
        decorators.timing(func)()
        self.assertEquals(func.call_count, 1)


class DelayFnTests(TestCase):

    def test_delay_fn(self):
        def double(x): return 2 * x
        decorated_function = decorators.delay_fn(seconds=0)(double)
        self.assertEquals(decorated_function(31), 62)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['djstarter',
 'djstarter.management',
 'djstarter.management.commands',
 'djstarter.migrations',
 'djstarter.tests']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2.15,<4.0.0',
 'coverage>=6.4.4,<7.0.0',
 'django-environ>=0.9.0,<0.10.0',
 'httpx[http2]>=0.23.0,<0.24.0',
 'jsonpickle>=2.1.0,<3.0.0',
 'psycopg2-binary>=2.9.2,<3.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'wheel>=0.36.2,<0.37.0']

setup_kwargs = {
    'name': 'dj-starter',
    'version': '0.1.15',
    'description': 'Django Project Bootstraper with Common Models and Utilities',
    'long_description': 'None',
    'author': 'Adrian',
    'author_email': 'adrian@rydeas.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/adrianmeraz/dj-starter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)
